from __future__ import annotations

from datetime import datetime, timedelta
from dataclasses import dataclass
from typing import Iterable

from agton.ton import Contract, Cell, Message, MessageRelaxed, begin_cell, Provider, TlbConstructor
from agton.ton import Builder
from agton.ton import Slice, Address
from agton.ton.crypto.signing import private_key_to_public_key

from agton.wallet.mnemonic import mnemonic_to_private_key, new_mnemonic

WALLET_V3R2_CODE = Cell.from_boc('b5ee9c720101010100710000deff0020dd2082014c97ba218201339cbab19f71b0ed44d0d31fd31f31d70bffe304e0a4f2608308d71820d31fd31fd31ff82313bbf263ed44d0d31fd31fd3ffd15132baf2a15144baf2a204f901541055f910f2a3f8009320d74a96d307d402fb00e8d101a4c8cb1fcb1fcbffc9ed54')
WALLET_V3R2_SUBWALLET_MAGIC = 698983191

@dataclass(frozen=True, slots=True)
class WalletV3R2Data(TlbConstructor):
    seqno: int
    subwallet: int
    public_key: bytes

    @classmethod
    def initial(cls, public_key: bytes, subwallet: int) -> WalletV3R2Data:
        return cls(
            seqno=0,
            subwallet=subwallet,
            public_key=public_key
        )
    
    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> WalletV3R2Data:
        seqno = s.load_uint(32)
        subwallet = s.load_uint(32)
        public_key = s.load_bytes(256 // 8)
        return cls(seqno, subwallet, public_key)

    def serialize_fields(self, b: Builder) -> Builder:
        return (
            b
            .store_uint(self.seqno, 32)
            .store_uint(self.subwallet, 32)
            .store_bytes(self.public_key)
        )


class WalletV3R2(Contract):
    def __init__(self,
                 address: Address,
                 private_key: bytes,
                 subwallet: int | None = None,
                 provider: Provider | None = None) -> None:
        if subwallet is None:
            subwallet = WALLET_V3R2_SUBWALLET_MAGIC + address.workchain
        self.subwallet = subwallet
        self.private_key = private_key
        super().__init__(address, provider)

    MessageWithMode = tuple[MessageRelaxed, int]

    def create_signed_external(self, 
                               messages_with_modes: Iterable[MessageWithMode],
                               valid_until: int,
                               seqno: int,
                               use_dummy_private_key: bool = False) -> Message:
        messages_with_modes = tuple(messages_with_modes)
        if len(messages_with_modes) > 4:
            raise ValueError('WalletV3R2 supports only up to 4 messages')
        packed_messages: Builder = Builder()
        for message, mode in messages_with_modes:
            packed_messages.store_uint(mode, 8)
            packed_messages.store_ref(message.to_cell())

        unsigned_body = (
            begin_cell()
            .store_uint(self.subwallet, 32)
            .store_uint(valid_until, 32)
            .store_uint(seqno, 32)
            .store_builder(packed_messages)
        )
        key = bytes([0] * 32) if use_dummy_private_key else self.private_key
        signature = unsigned_body.end_cell().sign(key)
        signed_body = (
            begin_cell()
            .store_bytes(signature)
            .store_builder(unsigned_body)
            .end_cell()
        )
        return self.create_external_message(signed_body)
    
    def _safety_check(self, mode: int, allow_dangerous: bool):
        if not (mode & 2) and not allow_dangerous:
            raise ValueError(
                'Sending message without SendIgnoreErrors flag set can be dangerous'
                'use alow_dangerous=True if you know what you doing, and want to suppress this error'
            )

    def send(self,
             m: MessageRelaxed,
             valid_until: int | None = None,
             mode: int = 3,
             *, 
             allow_dangerous: bool = False) -> bytes:
        self._safety_check(mode, allow_dangerous)
        if valid_until is None:
            t = datetime.now() + timedelta(minutes=3)
            valid_until = int(t.timestamp())
        signed_message = self.create_signed_external([(m, mode)], valid_until, self.seqno())
        return self.send_external_message(signed_message)
    
    def send_many(self,
                  messages_with_modes: Iterable[MessageWithMode],
                  valid_until: int | None = None, 
                  *, 
                  allow_dangerous: bool = False) -> bytes:
        messages_with_modes = tuple(messages_with_modes)
        for _, mode in messages_with_modes:
            self._safety_check(mode, allow_dangerous)

        if valid_until is None:
            t = datetime.now() + timedelta(minutes=3)
            valid_until = int(t.timestamp())
        signed_message = self.create_signed_external(messages_with_modes, valid_until, self.seqno())
        return self.send_external_message(signed_message)

    def seqno(self) -> int:
        s = self.run_get_method('seqno')
        match s:
            case int(): return s
            case _: raise TypeError(f'Unexpected result for seqno: {s!r}')

    @classmethod
    def from_private_key(cls,
                         private_key: bytes,
                         subwallet: int | None = None,
                         wc: int = 0,
                         provider: Provider | None = None) -> WalletV3R2:
        if subwallet is None:
            subwallet = WALLET_V3R2_SUBWALLET_MAGIC + wc
        public_key = private_key_to_public_key(private_key)
        data = WalletV3R2Data.initial(public_key, subwallet)
        address = cls.code_and_data_to_address(WALLET_V3R2_CODE, data.to_cell(), wc)
        return cls(address, private_key, subwallet, provider)

    @classmethod
    def from_mnemonic(cls,
                      mnemonic: str,
                      subwallet: int | None = None,
                      wc: int = 0,
                      provider: Provider | None = None) -> WalletV3R2:
        private_key = mnemonic_to_private_key(mnemonic)
        return cls.from_private_key(private_key, subwallet, wc, provider)

    @classmethod
    def create(cls,
               subwallet: int | None = None,
               wc: int = 0,
               provider: Provider | None = None) -> tuple[WalletV3R2, str]:
        mnemonic = new_mnemonic()
        return cls.from_mnemonic(mnemonic, subwallet, wc, provider), mnemonic



from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from agton.ton import Cell, Slice, MsgAddressInt, begin_cell
from agton.ton import Contract, HashmapCodec
from agton.ton.types import OutAction, ActionSendMsg, out_action
from agton.ton.types import OutList, out_list
from agton.ton.types.tlb import TlbConstructor, TlbDeserializationError
from agton.ton.cell.builder import Builder


@dataclass(frozen=True, slots=True)
class AddExtension(TlbConstructor):
    '''add_extension#02 addr:MsgAddressInt = W5ExtendedAction;'''
    addr: MsgAddressInt

    @classmethod
    def tag(cls):
        return 0x02, 8

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        addr = s.load_msg_address_int()
        return cls(addr)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_msg_address_int(self.addr)


@dataclass(frozen=True, slots=True)
class DeleteExtension(TlbConstructor):
    '''delete_extension#03 addr:MsgAddressInt = W5ExtendedAction;'''
    addr: MsgAddressInt

    @classmethod
    def tag(cls):
        return 0x03, 8

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        addr = s.load_msg_address_int()
        return cls(addr)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_msg_address_int(self.addr)


@dataclass(frozen=True, slots=True)
class SetSignatureAuthAllowed(TlbConstructor):
    '''set_signature_auth_allowed#04 allowed:Bool = W5ExtendedAction;'''
    allowed: bool

    @classmethod
    def tag(cls):
        return 0x04, 8

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        allowed = s.load_bool()
        return cls(allowed)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_bool(self.allowed)

W5ExtendedAction = AddExtension | DeleteExtension | SetSignatureAuthAllowed

def w5_extended_action(s: Slice) -> W5ExtendedAction:
    tag = s.preload_uint(8)
    if tag == AddExtension.tag()[0]:
        return AddExtension.deserialize(s)
    if tag == DeleteExtension.tag()[0]:
        return DeleteExtension.deserialize(s)
    if tag == SetSignatureAuthAllowed.tag()[0]:
        return SetSignatureAuthAllowed.deserialize(s)
    raise TlbDeserializationError(f'Unknown tag for ExtendedAction: {tag:08x}')


@dataclass(frozen=True, slots=True)
class ExtendedListLast(TlbConstructor):
    '''extended_list_last$_ action:W5ExtendedAction = W5ExtendedActionList 0;'''
    action: W5ExtendedAction

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        action = s.load_tlb(w5_extended_action)
        return cls(action)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_tlb(self.action)

@dataclass(frozen=True, slots=True)
class ExtendedListCons(TlbConstructor):
    '''extended_list_action$_ {m:#} action:W5ExtendedAction prev:^(W5ExtendedActionList m) = W5ExtendedActionList (m + 1);'''
    action: W5ExtendedAction
    prev: W5ExtendedActionList

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        action = s.load_tlb(w5_extended_action)
        prev = s.load_ref_tlb(w5_extended_action_list)
        return cls(action, prev)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_tlb(self.action)
        b.store_ref_tlb(self.prev)
        return b

W5ExtendedActionList = ExtendedListLast | ExtendedListCons

def w5_extended_action_list(s: Slice) -> W5ExtendedActionList:
    '''
    This is incorrect implementation
    will work as long as after W5ExtendedList in cell there are no possible refs
    '''
    if s.remaining_refs == 0:
        return ExtendedListLast.deserialize(s)
    return ExtendedListCons.deserialize(s)

@dataclass(frozen=True, slots=True)
class W5InnerRequest(TlbConstructor):
    '''
    w5_actions_request$_ {m:#} {n:#} 
        out_actions:(Maybe ^(OutList m)) 
        extended_actions:(Maybe (W5ExtendedActionList n)) 
        = W5InnerRequest m n;
    '''
    out_actions: OutList | None
    extended_actions: W5ExtendedActionList | None

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        out_actions = s.load_maybe_ref_tlb(out_list)
        extended_actions = s.load_maybe_tlb(w5_extended_action_list)
        return cls(out_actions, extended_actions)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_maybe_ref_tlb(self.out_actions)
        b.store_maybe_tlb(self.extended_actions)
        return b

@dataclass(frozen=True, slots=True)
class W5SignedRequest(TlbConstructor):
    '''
    w5_signed_request$_ {m:#} {n:#}
        wallet_id:    uint32
        valid_until:  uint32
        msg_seqno:    uint32
        inner:        (W5InnerRequest m n)
        signature:    bits512
    = W5SignedRequest m n;
    '''
    wallet_id: int
    valid_until: int
    msg_seqno: int
    inner: W5InnerRequest
    signature: bytes

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        wallet_id = s.load_uint(32)
        valid_until = s.load_uint(32)
        msg_seqno = s.load_uint(32)
        inner = s.load_tlb(W5InnerRequest)
        signature = s.load_bytes(64)
        return cls(wallet_id, valid_until, msg_seqno, inner, signature)

    def serialize_fields(self, b: Builder) -> Builder:
        raise NotImplementedError

    def __post_init__(self):
        if len(self.signature) != 64:
            raise ValueError(f'Expected 64 bytes in signature, but {len(self.signature)} found')

@dataclass(frozen=True, slots=True)
class W5InternalSignedRequest(TlbConstructor):
    '''w5_internal_signed_request#73696e74 {m:#} {n:#} request:(W5SignedRequest m n) = W5MsgBody m n;'''
    request: W5SignedRequest

    @classmethod
    def tag(cls):
        return 0x73696e74, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        request = s.load_tlb(W5SignedRequest)
        return cls(request)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_tlb(self.request)

@dataclass(frozen=True, slots=True)
class W5ExternalSignedRequest(TlbConstructor):
    '''w5_external_signed_request#7369676e {m:#} {n:#} request:(W5SignedRequest m n) = W5MsgBody m n;'''
    request: W5SignedRequest

    @classmethod
    def tag(cls):
        return 0x7369676e, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        request = s.load_tlb(W5SignedRequest)
        return cls(request)

    def serialize_fields(self, b: Builder) -> Builder:
        return b.store_tlb(self.request)

@dataclass(frozen=True, slots=True)
class W5ExtensionActionRequest(TlbConstructor):
    '''
    w5_extension_action_request#6578746e {m:#} {n:#}
        query_id:uint64 
        request:(W5InnerRequest m n) 
    = W5MsgBody m n;
    '''
    query_id: int
    request: W5InnerRequest

    @classmethod
    def tag(cls):
        return 0x6578746e, 32

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        query_id = s.load_uint(64)
        request = s.load_tlb(W5InnerRequest)
        return cls(query_id, request)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_uint(self.query_id, 64)
        b.store_tlb(self.request)
        return b

W5MsgBody = W5InternalSignedRequest | W5ExternalSignedRequest | W5ExtensionActionRequest

def w5_msg_body(s: Slice) -> W5MsgBody:
    tag = s.load_uint(64)
    if tag == W5InternalSignedRequest.tag()[0]:
        return W5InternalSignedRequest.deserialize(s)
    if tag == W5ExternalSignedRequest.tag()[0]:
        return W5ExternalSignedRequest.deserialize(s)
    if tag == W5ExtensionActionRequest.tag()[0]:
        return W5ExtensionActionRequest.deserialize(s)
    raise TlbDeserializationError(f'Unexpected tag for W5MsgBody: {tag:08x}')


@dataclass(frozen=True, slots=True)
class WalletV5Data(TlbConstructor):
    '''
    contract_state$_ 
        is_signature_allowed: bool
        seqno:                uint32
        wallet_id:            uint32
        public_key:           bits256
        extensions_dict:      (HashmapE 256 bool)
    = ContractState;
    '''
    is_signature_allowed: bool
    seqno: int
    wallet_id: int
    public_key: bytes
    extensions_dict: dict[bytes, bool]

    @classmethod
    def tag(cls):
        return None

    @classmethod
    def deserialize_fields(cls, s: Slice) -> Self:
        is_signature_allowed = s.load_bool()
        seqno = s.load_uint(32)
        wallet_id = s.load_uint(32)
        public_key = s.load_bytes(32)
        extensions_codec = HashmapCodec().with_bytes_keys(32).with_bool_values()
        extensions_hashmap_e = s.load_hashmap_e(256)
        extensions_dict = extensions_codec.decode(extensions_hashmap_e)
        return cls(is_signature_allowed, seqno, wallet_id, public_key, extensions_dict)

    def serialize_fields(self, b: Builder) -> Builder:
        b.store_bool(self.is_signature_allowed)
        b.store_uint(self.seqno, 32)
        b.store_uint(self.wallet_id, 32)
        b.store_bytes(self.public_key)
        extensions_codec = HashmapCodec().with_bytes_keys(32).with_bool_values()
        extensions_hashmap_e = extensions_codec.encode(self.extensions_dict)
        b.store_hashmap_e(extensions_hashmap_e, 256)
        return b
    
    @classmethod
    def initial(cls,
                public_key: bytes,
                wallet_id: int,
                is_signature_allowed: bool = True,
                extensions_dict: dict[bytes, bool] | None = None) -> WalletV5Data:
        if extensions_dict is None:
            extensions_dict = dict()
        return cls(
            is_signature_allowed=is_signature_allowed,
            seqno=0,
            wallet_id=wallet_id,
            public_key=public_key,
            extensions_dict=extensions_dict
        )

    def __post_init__(self):
        if len(self.public_key) != 32:
            raise ValueError(f'Expected 32 bytes in public_key, but {len(self.public_key)} found')
